	 title	 CPUID -- Determine CPU & NDP Type
	 page	 58,122
	 name	 CPUID

COMMENT|

CPUID purports to uniquely identify each Intel CPU & NDP used in IBM
PCs and compatibles.

Notes on Program Structure
--------------------------

   This program uses four segments, two classes, and one group.  It
demonstrates a useful technique for programmers who generate .COM
programs.  In particular, it shows how to use segment classes to
re-order segments, and how to eliminate the linker's warning message
about the absence of a stack segment.

   The correspondence between segments and classes is as follows:

	 Segment	Class
	 -------	-----
	 STACK		prog
	 DATA		data
	 MDATA		data
	 CODE		prog

   The segments appear in the above order in the program source to
avoid forward references in the CODE segment to labels in the
DATA/MDATA segments.  However, because the STACK segment appears first
in the file, it and all segments in the same class are made contiguous
by the linker.	Thus they precede the DATA/MDATA segments in the
resulting .COM file because the latter are in a different class.  In
this manner, although DATA and MDATA precede CODE in the source file,
their order is swapped in the .COM file.  That way there is no need
for an initial skip over the data areas to get to the CODE segment.
As a side benefit, declaring a STACK segment (as the first segment in
the source) also eliminates the linker's warning about that segment
being missing.	Finally, all segments are declared to be in the same
group so the linker can properly resolve offsets.

   Note that if you re-assemble the code for any reason, it is
important to use an assembler later than the IBM version 1.0.  That
version has a number of bugs including an annoying habit of
alphabetizing segment names in the .OBJ file.  Such gratuitous
behavior defeats the above technique as well as exhibits generally bad
manners.  If you use IBM MASM 2.0, be sure to specify /S to order the
segments properly.

   If the program reports results at variance with your knowledge of
the system, please contact the author.

Environments tested in:
		 CPU Speed
  System	  in MHz	CPU		 NDP
  ----------------------------------------------------------
  IBM PC	    4.77     Intel 8088 	Intel 8087-3
  IBM PC	    4.77     Intel 8088*	Intel 8087-3
  IBM PC XT	    4.77     Intel 8088 	none
  IBM PC XT	    4.77     Intel 8088 	Intel 8087-3
  IBM PC XT/286     6	     Intel 80286	none
  IBM PC AT	    6	     Intel 80286	Intel 80287
  IBM PC AT	    9	     Intel 80286	Intel 80287-8
  IBM PC AT	    6	     Intel 80286	none
  IBM PC AT	    8.5      Intel 80286	none
  IBM 3270 PC AT    6	     Intel 80286	none
  COMPAQ Portable   4.77     Intel 8088 	none
  COMPAQ Portable   4.77     NEC V20		none
  COMPAQ 286	    8	     Intel 80286	none
  COMPAQ 386	   16	     Intel 80386	Intel 80827-8
  AT&T PC 6300	    8	     Intel 8086 	Intel 8087-2
  AT&T PC 6300	    8	     NEC V30		Intel 8087-2
  TANDY 2000	    8	     Intel 80186	none
  HP 150	    4.77     Intel 8088 	none
  Zenith Z-160	    4.77     NEC V20		none
  HP Vectra RS/20  20	     Intel 80386	Intel 80387, Weitek 1167

  * = with faulty CPU

Program structure:
  Group PGROUP:
  Stack   segment STACK, byte-aligned, stack,  class 'prog'
  Program segment CODE,  word-aligned, public, class 'prog'
  Data	  segment DATA,  byte-aligned, public, class 'data'
  Data	  segment MDATA, byte-aligned, public, class 'data'

Assembly requirements:

  Use MASM 1.25 or later.
  With IBM's MASM 2.0 only, use /S to avoid
    alphabetizing the segment names.
  Use /r option to generate real NDP code.

  MASM CPUID/r; 		  to convert .ASM to .OBJ
  LINK CPUID;			  to convert .OBJ to .EXE
  EXE2BIN CPUID CPUID.COM	  to convert .EXE to .COM
  ERASE CPUID.EXE		  to avoid executing .EXE

  Note that the linker doesn't warn about a missing stack segment.

Copyright free.

Original code by:

  Bob Smith   May 1985	 bobs@access.digex.net

  Arthur Zachai suggested the technique to distinguish within the 808x
  and 8018x families by exploiting the difference in the length of
  their pre-fetch instruction queues.

Modifications by:

Who			When			Why
----------------------------------------------------------------------
Bob Smith		21 Jan 86		Distinguish NEC V20/V30s
						from Intel 8086/8088s

David E. Michener	28 Jun 86		Use /z to avoid INT 11h if
						on a machine which doesn't
						support interrupts below 20h

Bob Smith		15 Sep 86		Distinguish 286 from 386

Bob Smith		 6 Oct 86		Return length of PIQ

Dan Lewis		15 Dec 86		Use /r to allow NDP instructions
						on 286/386 machines w/o NDP

Bob Smith		17 Jun 87		Distinguish 287 from 387

Bob Smith		24 Nov 87		Detect Weitek 1167

Bob Smith		26 Apr 89		Detect 80486

Bob Smith &		15 Aug 91		Detect 80386 SX
Robert Hummel

Bob Smith		19 Aug 92		Detect CPUID instruction and
						Pentium CPU.
|

.386p
	 subttl  Structures, Records, Equates, & Macros
	 page
ARG_STR  struc

	 dw	 ?		; Caller's BP
ARG_OFF  dw	 ?		; Caller's offset
ARG_SEG  dw	 ?		;	   segment
ARG_FLG  dw	 ?		;	   flags

ARG_STR  ends

; Record to define bits in the CPU's & NDP's flags' registers

CPUFLAGS record  $ID:1,$VIP:1,$VIF:1,$AC:1,$VM:1,$RF:1, \
		 $R0:1,$NT:1,$IOPL:2,$OF:1,$DF:1,$IF:1,$TF:1,$SF:1,$ZF:1,$R1:1,$AF:1,$R2:1,$PF:1,$R3:1,$CF:1
NDPFLAGS record  $R4:3,$IC:1,$RC:2,$PC:2,$IEM:1,$R5:1,$PM:1,$UM:1,$OM:1,$ZM:1,$DM:1,$IM:1

CR0FLAGS record  $PG:1,$CD:1,$NW:1,$CR0RSV0:10,$AM:1,$CR0RSV1:1,$WP:1,\
		 $CR0RSV2:10,$NE:1,$ET:1,$TS:1,$EM:1,$MP:1,$PE:1
COMMENT|

FLG_PIQL	 Pre-fetch instruction queue length, 0 => 4-byte, 1 => 6-byte
FLG_08		 Intel 808x
FLG_NEC 	 NEC V20 or V30
FLG_18		 Intel 8018x
FLG_28		 Intel 8028x
FLG_38		 Intel 8038x
FLG_48		 Intel 8048x
FLG_38SX	 Intel 8038x SX
FLG_58		 Intel Pentium CPU

FLG_87		 Intel 8087
FLG_287 	 Intel 80287
FLG_387 	 Intel 80387

FLG_1167	 Weitek 1167

$FLG_CERR	 Faulty CPU
$FLG_NERR	 Faulty NDP switch setting

$FLG_CPUID	 CPU supports CPUID instruction

|

FLG_REC  record  $RSVD:5,$FLG_CPUID:1,$FLG_NERR:1,$FLG_CERR:1,$FLG_WTK:1,$FLG_NDP:3,$FLG_CPU:4

; CPU-related flags

FLG_PIQL equ	 0001b shl $FLG_CPU
FLG_08	 equ	 0000b shl $FLG_CPU
FLG_NEC  equ	 0010b shl $FLG_CPU
FLG_18	 equ	 0100b shl $FLG_CPU
FLG_28	 equ	 0110b shl $FLG_CPU
FLG_38	 equ	 1000b shl $FLG_CPU
FLG_48	 equ	 1010b shl $FLG_CPU
FLG_38SX equ	 1100b shl $FLG_CPU
FLG_58	 equ	 1110b shl $FLG_CPU

FLG_8088 equ	 FLG_08
FLG_8086 equ	 FLG_08 or FLG_PIQL
FLG_V20  equ	 FLG_NEC
FLG_V30  equ	 FLG_NEC or FLG_PIQL
FLG_80188 equ	 FLG_18
FLG_80186 equ	 FLG_18 or FLG_PIQL
FLG_80286 equ	 FLG_28 or FLG_PIQL
FLG_80386 equ	 FLG_38 or FLG_PIQL
FLG_80486 equ	 FLG_48 or FLG_PIQL
FLG_P5	  equ	 FLG_58 or FLG_PIQL

; NDP-related flags

FLG_NDPX equ	 000b shl $FLG_NDP	; Not present
FLG_NDPU equ	 001b shl $FLG_NDP	; Untested
FLG_87	 equ	 010b shl $FLG_NDP
FLG_287  equ	 011b shl $FLG_NDP
FLG_387  equ	 100b shl $FLG_NDP

FLG_1167 equ	 mask $FLG_WTK

FLG_CPUID equ	 mask $FLG_CPUID


BEL	 equ	 07h
TAB	 equ	 09h
LF	 equ	 0Ah
CR	 equ	 0Dh
EOS	 equ	 '$'


VECTOR	 struc

VOFF	 dw	 ?		; Offset
VSEG	 dw	 ?		; Segment

VECTOR	 ends


POPFF	 macro
	 local	 L1,L2

	 jmp	 short L2	; Skip over IRET
L1:
	 iret			; Pop the CS & IP pushed below along
				; with the flags, our original purpose
L2:
	 push	 cs		; Prepare for IRET by pushing current CS
	 call	 L1		; Push IP, jump to IRET

	 endm			; POPFF macro

TABLE	 macro	 TYP

	 push	 bx		; Save for a moment
	 and	 bx,mask $FLG_&TYP ; Isolate flags
	 mov	 cl,$FLG_&TYP	; Shift amount
	 shr	 bx,cl		; Shift to low-order
	 shl	 bx,1		; Times two to index table of words
	 mov	 dx,TYP&MSG_TAB[bx] ; DS:DX ==> descriptive message
	 pop	 bx		; Restore

	 mov	 ah,09h 	; Function code to display string at DS:DX
	 int	 21h		; Request DOS service

	 endm			; TABLE macro

REGSAVE  macro	 LIST		; Register save macro

	 irp	 XX,<LIST>
	 push	 XX
	 endm

	 endm			; REGSAVE

REGREST  macro	 LIST		; Register restore macro

	 irp	 XX,<LIST>
	 pop	 XX
	 endm

	 endm			; REGREST

DPALIGN  macro	 OFF,BYT	; Align up to para boundary
	 local	 BYT2

ifnb <BYT>
BYT2	 equ	 BYT
else
BYT2	 equ	 0
endif

	 ALIGNCOM OFF,BYT2,16

	 endm			; DPALIGN


ALIGNCOM macro	 OFF,BYT,N	; Common alignment macro
	 local	 LBL,M

LBL	 label	 byte		; Anchor ourselves
M	 equ	 ((BYT+LBL-OFF) mod N) ; Compute difference modulo alignment
if M				; If non-zero 'cause MASM doesn't handle 0 dup
	 db	 (N-M) dup (90h) ; Spit out NOPs
endif				; IF
	 endm			; ALIGNCOM

	 page
INT_VEC  segment use16 at 0	; Start INT_VEC segment

	 org	 4*01h
INT01_OFF dw	 ?		; Pointer to INT 01h
INT01_SEG dw	 ?

INT_VEC  ends			; End INT_VEC segment

PGROUP	 group	 STACK,CODE,DATA,MDATA

	 extrn	 VERS_H:abs,VERS_T:abs,VERS_U:abs

; The following segment both positions class 'prog' segments lower in
; memory than others so the first byte of the resulting .COM file is
; in the CODE segment, as well as satisfies the LINKer's need to have
; a stack segment.

STACK	 segment use16 byte stack 'prog' ; Start STACK segment
STACK	 ends			; End STACK segment

@OPCOD_INCBX equ 43h		; Opcode for INC BX
@OPCOD_NOP   equ 90h		; ...	     NOP

@BIT24	 equ	 1 shl 24	; Equate for Bit #24

I11_REC  record  $I11_PRN:2,$I11_RSV1:2,$I11_COM:3,$I11_RSV2:1,$I11_DISK:2,$I11_VID:2,$I11_RSV3:2,$I11_NDP:1,$I11_IPL:1

CPUFET_REC record $CPUFET_MSREG:1,  \
		  $CPUFET_TSCNT:1,  \
		  $CPUFET_4MBPG:1,  \
		  $CPUFET_IOBRK:1,  \
		  $CPUFET_NDPIN:1,  \
		  $CPUFET_ENHV86:1

CPUSIG_REC record $CPUSIG_MEM:4, \
		  $CPUSIG_FAM:4, \
		  $CPUSIG_MOD:4, \
		  $CPUSIG_REV:4

MODRM_REC record $MOD:2,$REG:3,$RM:3

DATA	 segment use16 dword public 'data' ; Start DATA segment
	 assume  ds:PGROUP

CPUID_LAST dd	 ?		; Highest CPUID function supported
CPUFET_FLAG dd	 ?		; CPU feature bits
CPUSIG_FLAG dd	 ?		; CPU signature bits
CR4_FLAG dd	 ?		; CR4 flags

OLDINT01_VEC label dword	; Save area for original INT 01h handler
OLDINT01_OFF dw  ?
OLDINT01_SEG dw  ?

NDP_ENV  dw	 7 dup (?)	; Save area for NDP environment

NDP_CW	 label	 word		; Save area for NDP control word
	 db	 ?
NDP_CW_HI db	 0		; High byte of control word
	 db	 15 dup (?)	; Room for the CPU to step on

DATA	 ends			; End DATA segment
	 subttl  Message Data Area
	 page
MDATA	 segment use16 dword public 'data' ; Start MDATA segment
	 assume  ds:PGROUP

MSG_START db	 'CPUID    -- Version '
	 db	 VERS_H,'.',VERS_T,VERS_U
	 db	 CR,LF,EOS

HEXTABLE db	 '0123456789ABCDEF' ; Hex translate table

MANUID	 db	 'The Pentium CPU manufacturer ID is "'
MANUID1  db	 12 dup (?),'".',CR,LF,EOS

MSG_CPUSIG db	 'The Pentium CPU ID is '
MSG_CPUSIG1 db	 '________.',CR,LF,EOS

MSG_CPUFET db	 'The Pentium CPU feature bits are '
MSG_CPUFET1 db	 '________.',CR,LF,EOS

MSG_CR4FLG db	 'The Pentium CPU CR4 value is '
MSG_CR4FLG1 db	 '________.',CR,LF,EOS

MSG_CPULST db	 'The highest CPUID function supported is '
MSG_CPULST1 db	 '________.',CR,LF,EOS

	 even
LCL_FLAGS dw	 0		; Local flags
@LCL_REAL equ	 8000h		; Use real code to detect 287 or 387 coprocessors
@LCL_I11H equ	 4000h		; Avoid INT 11h on machines which don't support it

LCL_OPTS db	 'rb'           ; Table of valid option letters
NLCL_OPTS equ	 $-LCL_OPTS	; # valid ...

	 even
LCL_ACTS dw	 PGROUP:CHECK_ARGS_REAL ; Table of corresponding actions
	 dw	 PGROUP:CHECK_ARGS_I11H

PIQL	 dw	 0		; Length of prefetch instruction queue
PIQL_CNT dw	 10		; Retry count to ensure valid PIQL

MSG_PIQL db	 'The length of the pre-fetch instruction queue is '
MSG_PIQL1 db	 '____h.',CR,LF,EOS

MSG_8088    db	 'CPU is an Intel 8088.',CR,LF,EOS
MSG_8086    db	 'CPU is an Intel 8086.',CR,LF,EOS
MSG_V20     db	 'CPU is an NEC V20.',CR,LF,EOS
MSG_V30     db	 'CPU is an NEC V30.',CR,LF,EOS
MSG_80188   db	 'CPU is an Intel 80188.',CR,LF,EOS
MSG_80186   db	 'CPU is an Intel 80186.',CR,LF,EOS
MSG_UNK2    db	 'CPU is an Intel 80286.',CR,LF,EOS
MSG_80286   db	 'CPU is an Intel 80286.',CR,LF,EOS
MSG_UNK3    db	 'CPU is an Intel 80386.',CR,LF,EOS
MSG_80386DX db	 'CPU is an Intel 80386 DX.',CR,LF,EOS
MSG_UNK4    db	 'CPU is an Intel 80486.',CR,LF,EOS
MSG_80486   db	 'CPU is an Intel 80486.',CR,LF,EOS
MSG_UNK5    db	 'CPU is an Intel 80386 SX.',CR,LF,EOS
MSG_80386SX db	 'CPU is an Intel 80386 SX.',CR,LF,EOS
MSG_UNK6    db	 'CPU is an Intel Pentium CPU.',CR,LF,EOS
MSG_P5	    db	 'CPU is an Intel Pentium CPU.',CR,LF,EOS

	 even
CPUMSG_TAB label word
	 dw	 PGROUP:MSG_8088	; 0000 = Intel 8088
	 dw	 PGROUP:MSG_8086	; 0001 = Intel 8086
	 dw	 PGROUP:MSG_V20 	; 0010 = NEC V20
	 dw	 PGROUP:MSG_V30 	; 0011 = NEC V30
	 dw	 PGROUP:MSG_80188	; 0100 = Intel 80188
	 dw	 PGROUP:MSG_80186	; 0101 = Intel 80186
	 dw	 PGROUP:MSG_UNK2	; 0110 = ?
	 dw	 PGROUP:MSG_80286	; 0111 = Intel 80286
	 dw	 PGROUP:MSG_UNK3	; 1000 = ?
	 dw	 PGROUP:MSG_80386DX	; 1001 = Intel 80386 DX
	 dw	 PGROUP:MSG_UNK4	; 1010 = ?
	 dw	 PGROUP:MSG_80486	; 1011 = Intel 80486
	 dw	 PGROUP:MSG_UNK5	; 1100 = ?
	 dw	 PGROUP:MSG_80386SX	; 1101 = Intel 80386 SX
	 dw	 PGROUP:MSG_UNK6	; 1110 = ?
	 dw	 PGROUP:MSG_P5		; 1111 = Intel Pentium CPU

	 even
NDPMSG_TAB label word
	 dw	 PGROUP:MSG_NDPX	; 000 = No NDP
	 dw	 PGROUP:MSG_NDPU	; 001 = NDP untested
	 dw	 PGROUP:MSG_8087	; 010 = Intel 8087
	 dw	 PGROUP:MSG_80287	; 011 = Intel 80287
	 dw	 PGROUP:MSG_80387	; 100 = Intel 80387

	 even
WTKMSG_TAB label word
	 dw	 PGROUP:MSG_EMPTY	; 0 = No Weitek 1167
	 dw	 PGROUP:MSG_1167	; 1 = Weitek 1167

MSG_NDPX db	 'NDP is not present.',CR,LF,EOS
MSG_NDPU db	 'NDP is untested.',CR,LF,EOS
MSG_8087 db	 'NDP is an Intel 8087.',CR,LF,EOS
MSG_80287 db	 'NDP is an Intel 80287.',CR,LF,EOS
MSG_80387 db	 'NDP is an Intel 80387.',CR,LF,EOS
MSG_1167 db	 'NDP is a Weitek 1167.',CR,LF,EOS
MSG_EMPTY db	 EOS

	 even
CERRMSG_TAB label word
	 dw	 PGROUP:MSG_CPUOK	; 0 = CPU healthy
	 dw	 PGROUP:MSG_CPUBAD	; 1 = CPU faulty

MSG_CPUOK db	 EOS			; No message
MSG_CPUBAD label byte
db BEL,'*** CPU incorrectly allows interrupts after a change to SS ***',CR,LF
db 'It should be replaced with a more recent version as it could crash the',CR,LF
db 'system at seemingly random times.',CR,LF,EOS

	 even
NERRMSG_TAB label word
	 dw	 PGROUP:MSG_NDPSWOK	; 0 = NDP switch set correctly
	 dw	 PGROUP:MSG_NDPSWERR	; 1 = NDP switch set incorrectly

MSG_NDPSWOK db	 EOS			; No message
MSG_NDPSWERR label  byte
db '*** The system board switch which indicates whether or not there is',CR,LF
db 'an NDP installed in the system is not properly set.  To correct this,',CR,LF
db 'flip switch 2 of switch block 1 on the system board.',CR,LF,EOS

MSG_REAL db	 '> Using real code to detect math coprocessor...',CR,LF,EOS
MSG_I11H db	 '> Bypassing INT 11h test of system board switch...',CR,LF,EOS

MSG_HELP db	 "CPUID.COM determines the CPU and NDP models in use.  In particular, it can",CR,LF
	 db	 "distinguish from each other any of the following CPUs and NDPs:",CR,LF
	 db	 CR,LF
	 db	 "   Intel 8086, 8087, 8088, 80186, 80188, 80286, 80287, 80386, 80387, 80486,",CR,LF
	 db	 "   Pentium CPU, NEC V20, V30, Weitek 1167",CR,LF
	 db	 CR,LF
	 db	 "It also reports on faulty 8086/8088s.  Such a CPU is faulty if it allows an",CR,LF
	 db	 "interrupt immediately after a change to the SS register.  If the program",CR,LF
	 db	 "indicates you have a faulty CPU, you should replace it.  The program also",CR,LF
	 db	 "checks on the system board's NDP switch in case there is an NDP installed.",CR,LF
	 db	 CR,LF
	 db	 "The program's syntax is as follows:",CR,LF
	 db	 "     CPUID     to run the program",CR,LF
	 db	 "     CPUID /r  to run the program using real code to detect a 287 or 387",CR,LF
	 db	 "     CPUID /b  to run the program but bypass the INT 11h test for correctly",CR,LF
	 db	 "                  set system board NDP switch",CR,LF
	 db	 CR,LF
	 db	 "If the program reports results at variance with your knowledge of the system,",CR,LF
	 db	 "please contact the author:",CR,LF
	 db	 "     Bob Smith -- bobs@access.digex.net",CR,LF
	 db	 EOS

MDATA	 ends			; End MDATA segment
	 subttl  Main Routine
	 page
CODE	 segment use16 word public 'prog' ; Start CODE segment
	 assume  cs:PGROUP
PSPSTART label	 byte
	 org	 100h		; Skip over PSP
INITIAL  proc	 near
	 assume  ds:PGROUP,es:PGROUP,ss:nothing

	 lea	 dx,MSG_START	; Starting message
	 mov	 ah,09h 	; Function code to display string at DS:DX
	 int	 21h		; Request DOS service

	 call	 CHECK_ARGS	; Check command line for parameters

	 call	 CPUID		; Check the CPU's identity
				; Return with BX = FLG_REC value
	 TABLE	 CPU		; Display CPU results
	 call	 DISP_CPUID	; Display CPUID results (if supported)
	 call	 DISP_P5	; Display Pentium CPU results (if present)
	 TABLE	 NDP		; Display NDP results
	 TABLE	 WTK		; Display Weitek results
	 TABLE	 CERR		; Display CPU ERR results
	 TABLE	 NERR		; Display NDP ERR results

	 mov	 ax,PIQL	; Get length of PIQ
	 lea	 di,MSG_PIQL1	; ES:DI ==> output save area
	 call	 DW2HEX 	; Convert AX to hex at ES:DI
	 lea	 dx,MSG_PIQL	; DS:DX ==> string to display
	 mov	 ah,09h 	; Function code to display string at DS:DX
	 int	 21h		; Request DOS service

	 mov	 ax,PIQL	; Get length of PIQ
	 mov	 ah,4Ch 	; Function call to exit with return code
	 int	 21h		; Return to DOS

	 assume  ds:nothing,es:nothing,ss:nothing

INITIAL  endp			; End INITIAL procedure
	 subttl  CHECK_ARGS Procedure
	 page
CHECK_ARGS proc  near		; Start CHECK_ARGS procedure
	 assume  ds:PGROUP,es:PGROUP,ss:nothing
COMMENT|

Check the command line for parameters.

The following parameters are allowed:

/r	 Use real code to detect 287 or 387 math coprocessor

|

	 REGSAVE <ax,dx,si>	; Save registers

	 mov	 si,81h 	; DS:SI ==> command line parameters
CHECK_ARGS_NEXT:
	 call	 SKIP_WHITE	; Skip over white space

	 lodsb			; Get next character

	 cmp	 al,CR		; Check for end-of-line
	 je	 short CHECK_ARGS_EXIT ; That's all folks

	 cmp	 al,'/'         ; Option separator?
	 jne	 short CHECK_ARGS_ERR ; No, give 'em help

	 lodsb			; Get option letter
	 or	 al,20h 	; Convert to lowercase

	 lea	 di,LCL_OPTS	; ES:DI ==> string of valid option letters
	 mov	 cx,NLCL_OPTS	; Length of above string
   repne scasb			; Search for it
	 jne	 short CHECK_ARGS_ERR ; Not present, give 'em help

	 sub	 di,1+offset es:LCL_OPTS ; Convert to origin-0
	 shl	 di,1		; Times two to index table of words

	 jmp	 LCL_ACTS[di]	; Take appropriate action

CHECK_ARGS_REAL:
	 lea	 dx,MSG_REAL	; DS:DX ==> message for this flag
	 mov	 ah,09h 	; Function code to display string at DS:DX
	 int	 21h		; Request DOS service

	 or	 LCL_FLAGS,@LCL_REAL ; Mark as present

	 jmp	 CHECK_ARGS_NEXT ; Go around again

CHECK_ARGS_I11H:
	 lea	 dx,MSG_I11H	; DS:DX ==> message for this flag
	 mov	 ah,09h 	; Function code to display string at DS:DX
	 int	 21h		; Request DOS service

	 or	 LCL_FLAGS,@LCL_I11H ; Mark as present

	 jmp	 CHECK_ARGS_NEXT ; Go around again

CHECK_ARGS_ERR:
	 lea	 dx,MSG_HELP	; DS:DX ==> help message
	 mov	 ah,09h 	; Function code to display string at DS:DX
	 int	 21h		; Request DOS service

	 mov	 ax,4CFFh	; Function call to exit with return code
	 int	 21h		; Return to DOS
CHECK_ARGS_EXIT:
	 REGREST <si,dx,ax>	; Restore

	 ret			; Return to caller

	 assume  ds:nothing,es:nothing,ss:nothing

CHECK_ARGS endp 		; End CHECK_ARGS procedure
	 subttl  SKIP_WHITE Procedure
	 page
SKIP_WHITE proc  near		; Start SKIP_WHITE procedure
	 assume  ds:PGROUP,es:PGROUP,ss:nothing
COMMENT|

Skip over white space.

On entry:

DS:SI	 ==>	 next character (possibly white space)

On exit:

DS:SI	 ==>	 next character not white space

|

	 lodsb			; Get next character

	 cmp	 al,' '         ; Check for blanks
	 je	 short SKIP_WHITE ; Go around again

	 cmp	 al,TAB 	; Check for TABs
	 je	 short SKIP_WHITE ; Go around again

	 dec	 si		; Back off to non-white character

	 ret			; Return to caller

	 assume  ds:nothing,es:nothing,ss:nothing

SKIP_WHITE endp 		; End SKIP_WHITE procedure
	 subttl  CPUID Procedure
	 page
CPUID	 proc	 near		; Start CPUID procedure
	 assume  ds:PGROUP,es:PGROUP,ss:nothing
COMMENT|

This procedure determines the type of CPU and NDP (if any) in use.

The possibilities include:

Intel 8086
Intel 8088
NEC   V20
NEC   V30
Intel 80186
Intel 80188
Intel 80286
Intel 80386
Intel 80486
Intel Pentium CPU
Intel 8087
Intel 80287
Intel 80387

   Also checked is whether or not the CPU allows interrupts after
changing the SS segment register.  If the CPU does, it is faulty and
should be replaced.

   Further, if an NDP is installed, non-AT machines should have a
system board switch set correspondingly.  Such a discrepancy is
reported upon.

   On exit, BX contains flag settings (as defined in $FLG record) which
the caller can check.  For example, to test for an Intel 80286, use

	 and	 bx,mask $FLAG_CPU

	 cmp	 bx,FLG_80286
	 je	 short ITSA286

|

	 REGSAVE <ax,cx,di,ds,es> ; Save registers

	 xor	 bx,bx		; Initialize flags

; Test for 80286/386/486/Pentium CPU -- these CPUs execute PUSH SP by first
; storing SP on stack, then decrementing it.  Earlier CPUs first decrement
; then store.

	 push	 sp		; Only 286 pushes pre-push SP
	 pop	 ax		; Get it back

	 cmp	 ax,sp		; Check for same
	 jne	 short CHECK_18x ; They aren't, try next class

	 call	 DIST_CPUS	; Distinguish a 286 from 386 from a 486 from
				; a Pentium CPU
	 jmp	 short CHECK_PIQL ; Join common code

; Test for 80186/80188 -- 18x and later CPUs mask shift/rotate operations
; mod 32; earlier CPUs use all 8 bits of CL.

CHECK_18x:
	 mov	 bx,FLG_18	; Assume it's an 8018x
	 mov	 cl,32+1	; 18x masks shift counts mod 32
				; Note we can't use just 32 in CL
	 mov	 al,0FFh	; Start with all bits set

	 shl	 al,cl		; Shift one position if 18x
	 jnz	 short CHECK_PIQL ; Some bits still on, so it's a 18x or later;
				; check PIQL

	 mov	 bx,FLG_NEC	; Assume it's an NEC V-series CPU
	 call	 CHECK_NEC	; See if it's an NEC chip
	 jcxz	 CHECK_PIQL	; Good guess, check PIQL

	 mov	 bx,FLG_08	; It's an 808x
COMMENT|

Check the length of the pre-fetch instruction queue (PIQ).

xxxx6 CPUs have a PIQ length of 6 bytes,
xxxx8 CPUs   "     "     "      4   "

Self-modifying code is used to distinguish the two PIQ lengths.

To overcome write pipelining in 286/386 chips, the largest value
over several executions of the subroutine is used.

|

CHECK_PIQL:
	 call	 PIQL_SUB	; Handled via subroutine

	 cmp	 cx,PIQL	; Use the larger of the two
	 jbe	 short CHECK_PIQL1 ; CX is smaller

	 mov	 PIQL,cx	; Save to report on later
CHECK_PIQL1:
	 dec	 PIQL_CNT	; One fewer times through the loop
	 jnz	 short CHECK_PIQL ; Jump if

	 cmp	 PIQL,4 	; Check PIQL
	 jbe	 short CHECK_ERR ; Jump if xxxx8

	 or	 bx,FLG_PIQL	; PIQ length is 5 or longer

; Test for faulty chip (allows interrupts after change to SS register)

CHECK_ERR:
	 call	 ERR_SUB	; Handled via subroutine
	 jcxz	 CHECK_NDP	; If CX is 0, the DEC was executed,
				; and the CPU is OK
	 or	 bx,mask $FLG_CERR ; It's a faulty chip
CHECK_NDP:
	 call	 NDP_SUB	; Handled via subroutine

	 REGREST <es,ds,di,cx,ax> ; Restore registers
	 assume  ds:nothing,es:nothing

	 ret			; Return to caller

	 assume  ds:nothing,es:nothing,ss:nothing

CPUID	 endp			; End CPUID procedure
	 subttl  Distinguish A 286 From 386 From A 486 From A Pentium CPU
	 page
DIST_CPUS proc near
	 assume  ds:PGROUP,es:PGROUP,ss:nothing
COMMENT|

The test for 286 vs. 386/486 is done by attempting to set flag bits in
the high-order nibble of the flag word.  If that's successful, it's a
386/486; otherwise, it's a 286.

The test for 386 vs. 486 is done by attempting to set the Alignment Check
bit in the high-order word of the extended flag word.  If that's successful,
it's a 486; otherwise, it's a 386.

The test for 486 vs. Pentium CPU is done by testing for the CPUID instruction
and if it's supported, checking the CPU signature for family value of
5 or greater.

|

	 REGSAVE <ax>		; Save register

	 pushf			; Save flags for a moment

	 mov	 ax,0F000h	; Try to set high bits in flag register

	 push	 ax		; Move into flag register
	 popf

	 pushf			; Get flags back into AX
	 pop	 ax

	 popf			; Restore original flags

	 test	 ax,0F000h	; Any bits set?
	 jz	 short ITSA286	; No, so it's a 286

; Distinguish a 386 from a 486

	 call	 DIST_386or486	; Tell 'em apart
	 jc	 short ITSA486	; Jump if it's a 486

; Distinguish a 386DX from a 386SX

	 call	 DIST_386DXorSX ; Tell 'em apart
	 jc	 short ITSA386_DX  ; Jump if it's a 386 DX

	 or	 bx,FLG_38SX	; It's a 38x SX

	 jmp	 short DIST_CPUS_EXIT ; Join common exit code

ITSA386_DX:
	 or	 bx,FLG_38	; It's a 38x DX

	 jmp	 short DIST_CPUS_EXIT ; Join common exit code

ITSA486:

; Determine whether or not this CPU supports the CPUID instruction

	 call	 IZIT_CPUID	; Duzit support the CPUID instruction?
	 jnc	 short @F	; Jump if not

	 or	 bx,FLG_CPUID	; It does
@@:

; Distinguish a 486 from a Pentium CPU

	 call	 DIST_486orP5	; Tell 'em apart
	 jc	 short ITSAP5	; Jump if it's a Pentium CPU

	 or	 bx,FLG_48	; It's a 48x

	 jmp	 short DIST_CPUS_EXIT ; Join common exit code

ITSAP5:
	 or	 bx,FLG_58	; It's a Pentium CPU

	 jmp	 short DIST_CPUS_EXIT ; Join common exit code

ITSA286:
	 or	 bx,FLG_28	; It's a 28x
DIST_CPUS_EXIT:
	 REGREST <ax>		; Restore

	 ret			; Return to caller

	 assume  ds:nothing,es:nothing,ss:nothing

DIST_CPUS endp			; End DIST_CPUS procedure
	 subttl  Distinguish A 386 From 486
	 page
DIST_386or486 proc near
	 assume  ds:PGROUP,es:PGROUP,ss:nothing
COMMENT|

The test for 386 vs. 486 is done by attempting to set the Alignment Check
bit in the high-order word of the extended flag word.  If that's successful,
it's a 486; otherwise, it's a 386.

On exit:

CF	 =	 0 if it's a 386
	 =	 1 if it's a 486

|

	 push	 ebp		; Save to align stack
	 mov	 ebp,esp	; Save original stack pointer

	 and	 esp,0FFFFh and (not (4-1)) ; Align on a dword boundary in case
				; we're in V86 mode and $AM is set in CR0
				; Also clear high-order of ESP as we're using
				; it as an address register
	 push	 eax		; Save to use as scratch
	 pushfd 		; Save original flags

	 pushfd 		; Save original flags
	 or	 dword ptr [esp],mask $AC ; Set Alignment Check bit
	 popfd			; Put into effect

	 pushfd 		; Put onto the stack
	 pop	 eax		; Copy to register

	 test	 eax,mask $AC	; Izit still set?
	 jnz	 short @F	; Yes, so it's a 486

	 popfd			; Restore original flags

	 clc			; Indicate it's a 386

	 jmp	 short DIST_386or486_EXIT ; Join common code

@@:
	 popfd			; Restore original flags

	 stc			; Indicate it's a 486
DIST_386or486_EXIT:
	 pop	 eax		; Restore

	 mov	 esp,ebp	; Restore original pointer
	 pop	 ebp		; Restore

	 ret			; Return to caller

	 assume  ds:nothing,es:nothing,ss:nothing

DIST_386or486 endp		; End DIST_386or486 procedure
	 subttl  Distinguish A 386 DX From 386 SX
	 page
DIST_386DXorSX proc near
	 assume  ds:PGROUP,es:PGROUP,ss:nothing
COMMENT|

The test for 386 DX vs. 386 SX is done by attempting to set the ET bit
in CR0.  If it succeeds, the CPU is a DX; if it fails, it's an SX.

On exit:

CF	 =	 0 if it's a 386 SX
	 =	 1 if it's a 386 DX

|

	 REGSAVE <eax,ebx>	; Save for a moment

	 mov	 eax,cr0	; Get control register #0
	 mov	 ebx,eax	; Save to restore later

	 and	 eax,not (mask $ET) ; ET=0 works on DX only
	 mov	 cr0,eax	; Tell the CPU about it
	 mov	 eax,cr0	; Read it back in
	 mov	 cr0,ebx	; Restore original value

	 test	 eax,mask $ET	; Izit sticky?
	 jnz	 short DIST_386DXorSX_EXIT ; Jump if so (note CF=0)

	 stc			; Indicate it's a 386 DX
DIST_386DXorSX_EXIT:
	 REGREST <ebx,eax>	; Restore

	 ret			; Return to caller

	 assume  ds:nothing,es:nothing,ss:nothing

DIST_386DXorSX endp		; End DIST_386DXorSX procedure
	 subttl  Distinguish A 486 From Pentium CPU
	 page
DIST_486orP5 proc near
	 assume  ds:PGROUP,es:PGROUP,ss:nothing
COMMENT|

The test for 486 vs. Pentium CPU is done by testing for the CPUID instruction
and if it's supported, checking the CPU signature for family value of
5 or greater.

On exit:

CF	 =	 0 if it's a 486
	 =	 1 if it's a Pentium CPU

|

	 test	 bx,FLG_CPUID	; Duzit support this instruction?
	 jz	 short DIST_486orP5_EXIT ; Jump if not (note CF=0)

	 REGSAVE <eax,ebx,ecx,edx> ; Save for a moment

	 mov	 eax,0		; Function code to retrieve manufacture ID
;;;;;;;; CPUID			; Return with EAX = highest CPUID function value
	 db	 0Fh,0A2h	;	      EBX,EDX,ECX = manufacturer ID

	 mov	 CPUID_LAST,eax ; Save for later use
	 mov	 dword ptr MANUID1[0],ebx ; Save for later use
	 mov	 dword ptr MANUID1[4],edx ; ...
	 mov	 dword ptr MANUID1[8],ecx ; ...

	 mov	 eax,1		; Function code to retrieve feature bits
;;;;;;;; CPUID			; Return with EAX = stepping info
	 db	 0Fh,0A2h	;	      EBX, ECX reserved
				;	      EDX = feature bits

	 mov	 CPUSIG_FLAG,eax ; Save CPU signature bits
	 mov	 CPUFET_FLAG,edx ; Save feature bits

	 and	 eax,mask $CPUSIG_FAM ; Isolate the family field
	 shr	 eax,$CPUSIG_FAM ; Shift to low-order

	 cmp	 al,05h 	; Izit Pentium CPU or later?
	 jb	 short DIST_486orP5_XP5 ; Jump if not (note CF=1)

;;;;;;;; mov	 eax,cr4	; Get CPU extensions register
	 db	 0Fh,20h,(11b shl $MOD) or (4 shl $REG) or (000b shl $RM) ; ...
	 mov	 CR4_FLAG,eax	; Save for later use

	 clc			; Ensure CF=0
DIST_486orP5_XP5:
	 REGREST <edx,ecx,ebx,eax> ; Restore

	 cmc			; Complement to CF=1 if it's a Pentium CPU
				;		  =0 if not
DIST_486orP5_EXIT:
	 ret			; Return to caller

	 assume  ds:nothing,es:nothing,ss:nothing

DIST_486orP5 endp		; End DIST_486orP5 procedure
	 subttl  Determine Support of CPUID Instruction
	 page
IZIT_CPUID proc near
	 assume  ds:PGROUP,es:PGROUP,ss:nothing
COMMENT|

The test for the CPUID instruction is done by attempting to set the ID
bit in the high-order word of the extended flag dword.	If that's
successful, the CPUID instruction is supported; otherwise, it's not.

On exit:

CF	 =	 1 if it's supported
	 =	 0 otherwise

|

	 push	 bp		; Save to address the stack
	 clc			; Assume it's not supported
	 pushfd 		; Save original flags
	 pushfd 		; Save temporary flags

IZIT_CPUID_STR struc

IZIT_CPUID_TMPEFL dd ?		; Temporary EFL
IZIT_CPUID_RETEFL dd ?		; Return EFL
	 dw	 ?		; Caller's BP

IZIT_CPUID_STR ends

	 mov	 bp,sp		; Address the stack
	 or	 [bp].IZIT_CPUID_TMPEFL,mask $ID ; Set ID bit
	 popfd			; Put into effect

	 pushfd 		; Put back onto the stack to test

	 test	 [bp].IZIT_CPUID_TMPEFL,mask $ID ; Izit still set?
	 jz	 short @F	; No, so it's not supported

	 or	 [bp].IZIT_CPUID_RETEFL,mask $CF ; Indicate it's supported
@@:
	 popfd			; Restore temporary flags
	 popfd			; Restore original flags
	 pop	 bp		; Restore

	 ret			; Return to caller

	 assume  ds:nothing,es:nothing,ss:nothing

IZIT_CPUID endp 		; End IZIT_CPUID procedure
	 subttl  Check For NEC V20/V30
	 page
CHECK_NEC proc	 near
	 assume  ds:PGROUP,es:PGROUP,ss:nothing
COMMENT|

   The NEC V20/V30 CPUs are very compatible with the Intel 8086/8088.
The only point of "incompatiblity" is that they do not contain a bug
found in the Intel CPUs.  Specifically, the NEC CPUs correctly restart
an interrupted multi-prefix string instruction at the start of the
instruction.  The Intel CPUs incorrectly restart it in the middle of
the instruction.  This routine tests for that situation by executing
such an instruction for a sufficiently long period of time for a timer
interrupt to occur.  If at the end of the instruction, CX is zero,
it must be an NEC CPU; if not, it's an Intel CPU.

   Note that we're counting on the timer interrupt to do its thing
every 18.2 times per second.

   Here's a worst case analysis:  An Intel 8086/8088 executes 65535
iterations of LODSB ES:[SI] in 2+9+13*65535 = 851,966 clock ticks.  If
the Intel 8086/8088 is running at 15 MHz, each clock tick is 66.67
nanoseconds, hence the entire operation takes 56.8 milliseconds.  If the
timer is running at normal speed, it interrupts the CPU every 55
millseconds and so should interrupt the repeated string instruction at
least once.

|

	 mov	 cx,0FFFFh	; Move a lot of data
	 sti			; Ensure timer enabled

; Execute multi-prefix instruction.  Note that the value of ES as
; well as the direction flag setting is irrelevant.

	 push	 ax		; Save registers
	 push	 si
     rep lods	 byte ptr es:[si]
	 pop	 si		; Restore
	 pop	 ax

; On exit, if CX is zero, it's an NEC CPU, otherwise it's an Intel CPU

	 ret			; Return to caller

	 assume  ds:nothing,es:nothing,ss:nothing

CHECK_NEC endp
	 subttl  Pre-fetch Instruction Queue Subroutine
	 page
PIQL_SUB proc	 near
	 assume  ds:PGROUP,es:PGROUP,ss:nothing
COMMENT|

   This subroutine attempts to discern the length of the CPU's
pre-fetch instruction queue (PIQ).

   It stores a new instruction into the instruction stream
following the STOSB.  The loop proceeds backwards from the end
of the stream to the beginning.  At the point the inserted
instruction is not executed, we have found the last byte in
the PIQ.  The value in CX at that time is then the PIQ length.

|

	 REGSAVE <ax,bx,dx,si,di> ; Save registers

@REP	 equ	 1024		; Maximum length of PIQ
				; we can handle

	 std			; Store backwards

	 mov	 cx,@REP	; Loop counter
	 lea	 di,LAB_NOP+@REP-1 ; ES:DI ==> PIQL last byte
				; in fill area

	 mov	 al,@OPCOD_INCBX ; Change to INC BX
	 mov	 ah,@OPCOD_NOP	; Save a NOP here to restore
	 mov	 si,1		; Divisor
	 xor	 dx,dx		; Zero high-order word for divide
	 cli			; Ensure interrupts are disabled, otherwise
				; a timer tick could disturb the PIQ filling

; Ensure that the instruction following the STOSB is para-aligned
; The constant 11 is equal to @DIFF but MASM won't let us use
; a foward reference.

	 DPALIGN PSPSTART,11	; Align it
PIQL_SUB_NEXT:
	 xor	 bx,bx		; Initialize counter
	 div	 si		; Mark time to fill the PIQ
	 div	 si		; Mark time to fill the PIQ
	 div	 si		; Mark time to fill the PIQ
	 div	 si		; Mark time to fill the PIQ
	 stosb			; Change the instruction

; The PIQ begins filling here

LAB_NOP  label	 byte
	 rept	 @REP
	 nop			;; Fill byte
	 endm

@DIFF	 equ	 LAB_NOP-PIQL_SUB_NEXT ; Code length to precede para-alignment

if (PSPSTART-LAB_NOP) mod 16
%OUT Misalignment at LAB_NOP!!
endif

	 mov	 es:[di+1],ah	; Restore the NOP

	 and	 bx,bx		; Did we execute it?
	 jz	 short @F	; Jump if not

	 dec	 cx		; One fewer execution
	 jz	 short @F	; Jump if we're done

	 jmp	 PIQL_SUB_NEXT	; Go around again

@@:
	 sti			; Restore interrupts
	 cld			; Restore direction flag

	 REGREST <di,si,dx,bx,ax> ; Restore

; At the end, CX has the length of the PIQ

	 ret			; Return to caller

	 assume  ds:nothing,es:nothing,ss:nothing

PIQL_SUB endp			; End PIQL_SUB procedure
	 subttl  Check For Faulty Interrupts
	 page
ERR_SUB  proc	 near
	 assume  ds:PGROUP,es:PGROUP
COMMENT|

Test for faulty chip (allows interrupts after change to SS register).
Setup a handler for INT 01h (single-step interrupt) and turn on that
flag just before executing a POP SS.  If the CPU allows a single-step
interrupt after the POP SS, it's faulty.

On exit:

CX	 =	 1 if CPU is faulty
	 =	 0 if OK

|

	 REGSAVE <ax,bp,ds>	; Save registers

	 mov	 bp,sp		; Save to restore later

	 mov	 ax,seg INT_VEC ; Prepare to address interrupt vector segment
	 mov	 ds,ax		; DS points to segment 0
	 assume  ds:INT_VEC	; Tell the assembler

	 cli			; Nobody move while we swap

	 lea	 ax,INT01	; Point to our own handler
	 xchg	 ax,INT01_OFF	; Get and swap offset
	 mov	 OLDINT01_OFF,ax ; Save to restore later

	 mov	 ax,cs		; Our handler's segment
	 xchg	 ax,INT01_SEG	; Get and swap segment
	 mov	 OLDINT01_SEG,ax ; Save to restore later

; Note we continue with interrupts disabled to avoid an external interrupt
; occurring during this test.

	 mov	 cx,1		; Initialize a register
	 push	 ss		; Save SS to store back into itself

	 pushf			; Move flags
	 pop	 ax		; ...into AX
	 or	 ax,mask $TF	; Set trap flag
	 push	 ax		; Place onto stack
	 POPFF			; ...and then into effect
				; Some CPUs effect the trap flag immediately,
				;   some wait one instruction.
	 nop			; Allow interrupt to take effect
POST_NOP:
	 pop	 ss		; Change the stack segment register (to itself)
	 dec	 cx		; Normal CPUs execute this instruction before
				; recognizing the single-step interrupt

	 jmp	 short INT01_DONE ; Join common code

INT01:

; Note IF=TF=0

; If we're stopped at or before POST_NOP, continue on

	 push	 bp		; Prepare to address the stack
	 mov	 bp,sp		; Hello, Mr. Stack

	 cmp	 [bp].ARG_OFF,offset cs:POST_NOP ; Check offset
	 pop	 bp		; Restore
	 ja	 short INT01_DONE ; We're done

	 iret			; Return to caller

INT01_DONE:

; Restore old INT 01h handler

	 mov	 ax,OLDINT01_VEC.VOFF ; Get old INT 01h handler offset
	 mov	 INT01_OFF,ax	; Restore

	 mov	 ax,OLDINT01_VEC.VSEG ; ...			segment
	 mov	 INT01_SEG,ax	; Restore

	 sti			; Allow interrupts again (IF=1)

	 mov	 sp,bp		; Strip IP, CS, and Flags from stack

	 REGREST <ds,bp,ax>	; Restore
	 assume  ds:PGROUP	; Tell the assembler

	 ret			; Return to caller

	 assume  ds:nothing,es:nothing,ss:nothing

ERR_SUB  endp			; End ERR_SUB procedure
	 subttl  Check For Numeric Data Processor
	 page
NDP_SUB  proc	 near
	 assume  ds:PGROUP,es:PGROUP,ss:nothing
COMMENT|

   Test for a Numeric Data Processor -- Intel 8087, 80287, or 80387.
An 8087 allows FDISI, an 80287/80387 ignores it.  The 80287 and 80387
can be distinguished through their different treatment of the infinity
closure setting.

   In general, the technique used is passive -- it leaves the NDP in
the same state in which it is found.

   Unfortunately, some IBM PC/ATs and 3270/ATs without an NDP don't
handle floating-point instructions correctly.  In particular, when
no-WAIT NDP instruction is executed on those systems, they wipe out
the memory location and all bytes following in the same segment.  To
overcome this bug, Dan Lewis has suggested a technique which computes
the segment and offset of the location into which the store is made
such that the offset is in the last paragraph of the segment.  This
way, the wipe out is harmless.

On exit:

BX	 =	 $FLG_NDP & $FLG_NERR bits set as appropriate.

|

	 REGSAVE <ax,cx,di>	; Save registers

	 call	 CHECK_1167	; See if there's a Weitek 1167 in the system

; Because some IBM PC/ATs and 3270/ATs without an NDP don't handle
; floating-point instructions correctly, we check for a 286 explicitly
; and rely upon the equipment flags to tell us if there's an NDP installed.
; This behavior is also present on some 386s.

	 test	 LCL_FLAGS,@LCL_REAL ; Use real code or not?
	 jnz	 short NDP_SUB1 ; It's real

	 mov	 ax,bx		; Copy CPUID bits for destructive testing
	 and	 ax,(mask $FLG_CPU) and not FLG_PIQL ; Isolate CPU bits

	 cmp	 ax,FLG_28	; Izit a 28x?
	 je	 short NDP_SUB0 ; Yes, skip NDP instructions

	 cmp	 ax,FLG_38	; Izit a 38x?
	 je	 short NDP_SUB0 ; Yes, skip NDP instructions

	 cmp	 ax,FLG_48	; Izit a 48x?
	 jne	 short NDP_SUB1 ; Not this time
NDP_SUB0:
	 test	 LCL_FLAGS,@LCL_I11H ; Skip INT 11h test?
	 jnz	 short NDP_SUB_UN ; Yes

	 int	 11h		; Get equipment flags into AX

	 test	 ax,mask $I11_NDP ; Check NDP-installed bit
	 jz	 short NDP_SUB_EXIT0 ; Not installed

	 call	 DIST_287or387	; Distinguish a 287 from a 387
NDP_SUB_EXIT0:
	 jmp	 NDP_SUB_EXIT	; Join common exit code

NDP_SUB_UN:
	 or	 bx,FLG_NDPU	; Mark as untested

	 jmp	 NDP_SUB_EXIT	; Join common exit code

NDP_SUB1:

; Unfortunately, some systems don't handle NDP instructions well
; even without the WAIT opcode if there's no NDP installed.  Thus
; if the system board switch is mis-set, we could lockup the system
; with the FNSTENV instruction if we attempt it.

;;;;;;;; push	 es		; Save for a moment
;;;;;;;;
;;;;;;;; lea	 di,NDP_ENV+(size NDP_ENV)-1 ; Offset of end of environment
;;;;;;;; call	 MAX_OFFSET	; Return with ES:DI ==> NDP_ENV and DI largest
;;;;;;;; assume  es:nothing
;;;;;;;; sub	 di,(size NDP_ENV)-1 ; Back off to start of NDP_ENV
;;;;;;;;
;;;;;;;; cli			; Protect FNSTENV
;;;;;;;; fnstenv es:[di]	; If NDP present, save current environment,
;;;;;;;;			; otherwise, this instruction is ignored
;;;;;;;; sti			; Allow interrupts
;;;;;;;;
;;;;;;;; pop	 es		; Restore
;;;;;;;; assume  es:PGROUP
;;;;;;;;
;;;;;;;; mov	 cx,50/7	; Cycle this many times
;;;;;;;; loop	 $		; Wait for result to be stored
;;;;;;;;
	 fninit 		; Initialize processor to known state
	 jmp	 short $+2	; Wait for initialization

	 push	 es		; Save for a moment

	 lea	 di,NDP_CW+(size NDP_CW)-1 ; Offset of end of control word
	 call	 MAX_OFFSET	; Return with ES:DI ==> NDP_CW and DI largest
	 assume  es:nothing
	 sub	 di,(size NDP_CW)-1 ; Back off to start of control word

	 fnstcw  es:[di]	; Save control word

	 pop	 es		; Restore
	 assume  es:PGROUP

	 jmp	 short $+2	; Wait for result to be stored
	 jmp	 short $+2

	 and	 NDP_CW,not mask $IC ; Turn off infinity control in case of 387

	 cmp	 NDP_CW_HI,03h	; Check for NDP initial control word
	 jne	 short NDP_SUB_NONE ; No NDP installed

	 test	 LCL_FLAGS,@LCL_I11H ; Skip INT 11h test?
	 jnz	 short NDP_SUB2 ; Yes

	 int	 11h		; Get equipment flags into AX

	 test	 ax,mask $I11_NDP ; Check NDP-installed bit
	 jnz	 short NDP_SUB2 ; It's correctly set

	 or	 bx,mask $FLG_NERR ; Mark as in error
NDP_SUB2:
	 and	 NDP_CW,not mask $IEM ; Enable interrupts (IEM=0, 8087 only)
	 fldcw	 NDP_CW 	; Reload control word
	 fdisi			; Disable interrupts (IEM=1) on 8087,
				; ignored by 80287/80387
	 fstcw	 NDP_CW 	; Save control word
;;;;;;;; fldenv  NDP_ENV	; Restore original NDP environment
;;;;;;;;			; No need to wait for environment to be loaded

	 test	 NDP_CW,mask $IEM ; Check Interrupt Enable Mask (8087 only)
	 jnz	 short NDP_SUB_8087 ; It changed, hence NDP is an 8087

	 call	 DIST_287or387	; Distinguish a 287 from a 387

	 jmp	 short NDP_SUB_EXIT ; Exit with flags in BX

NDP_SUB_8087:
	 or	 bx,FLG_87	; NDP is an 8087

	 jmp	 short NDP_SUB_EXIT ; Join common exit code

NDP_SUB_NONE:
	 test	 LCL_FLAGS,@LCL_I11H ; Skip INT 11h test?
	 jnz	 short NDP_SUB_EXIT ; Yes

	 int	 11h		; Get equipment flags into AX

	 test	 ax,mask $I11_NDP ; Check NDP-installed bit
	 jz	 short NDP_SUB_EXIT ; It's correctly set

	 or	 bx,mask $FLG_NERR ; Mark as in error
NDP_SUB_EXIT:
	 REGREST <di,cx,ax>	; Restore

	 ret			; Return to caller

	 assume  ds:nothing,es:nothing,ss:nothing

NDP_SUB  endp			; End NDP_SUB procedure
	 subttl  Check For Weitek 1167
	 page
CHECK_1167 proc near
	 assume  ds:PGROUP,es:PGROUP,ss:nothing
COMMENT|

See if there's a Weitek 1167 in the system.

To determine that, clear EAX, call INT 11h, and test bit 24
in EAX.  If set, the coprocessor is present; if not, then
it's not.  Obviously, we must be running on a 386.

|

	 test	 LCL_FLAGS,@LCL_I11H ; Skip INT 11h test?
	 jnz	 short CHECK_1167_EXIT ; Yes, 1167 untested

	 cmp	 bx,FLG_38	; Are we on a 38x?
	 jne	 short CHECK_1167_EXIT ; No, thus no 1167

	 push	 eax		; Save for a moment

	 xor	 eax,eax	; Clear entire register
	 int	 11h		; Get equipment flags

	 test	 eax,@BIT24	; Test bit 24
	 jz	 short CHECK_1167_EXIT0 ; Not present

	 or	 bx,FLG_1167	; Mark as present
CHECK_1167_EXIT0:
	 pop	 eax		; Restore
CHECK_1167_EXIT:
	 ret			; Return to caller

	 assume  ds:nothing,es:nothing,ss:nothing

CHECK_1167 endp 		; End CHECK_1167 procedure
	 subttl  Distinguish A 287 From 387
	 page
DIST_287or387 proc near
	 assume  ds:PGROUP,es:PGROUP,ss:nothing
COMMENT|

   Distinguish a 287 from a 387.

   Both the 80287 and 80387 are initialized with the infinity closure
bit set to one.  However, only the 80287 is sensitive to the value of
this bit.  Ordinarily when this bit is set to one, the chip uses
projective closure; when it is cleared to zero, the chip uses affine
closure.  Thus the 80287 is initialized to use projective closure, but
that state can be changed through the infinity closure bit in the
control word.  On the other hand, the 80387 is initialized to use
affine closure and remains in that state independent of the setting of
the infinity closure bit.  The two NDPs can be distinguished by
executing code which is sensitive to the setting of the infinity
closure bit.

   The algorithm used is based upon one published by Intel on how to
detect the 80387.

On exit:

BX	 =	 $FLG_NDP bits set as appropriate.

|

.287
	 fstenv  NDP_ENV	; Save current environment
	 finit			; Initialize processor to known state
				; The 80287 is using projective
				; closure for arithmetic, the 80387 is
				; using affine closure

	 fld1			; Generate infinity
	 fldz			;  by dividing zero into one
	 fdiv			; ST0 = +infinity
	 fld	 st(0)		; Copy it
	 fchs			; ST0 = -infinity, ST1 = +infinity
	 fcompp 		; Compare them and pop both from stack
	 fstsw	 ax		; Get status word
	 fldenv  NDP_ENV	; Restore original NDP environment
	 sahf			; Copy into flags
	 jz	 short DIST_287or387_PROJ ; Jump if the two are equal
				; (projective closure)

	 or	 bx,FLG_387	; NDP is an 80387

	 jmp	 short DIST_287or387_EXIT ; Join common exit code

DIST_287or387_PROJ:
	 or	 bx,FLG_287	; NDP is an 80287
DIST_287or387_EXIT:
	 ret			; Return to caller
.8087
	 assume  ds:nothing,es:nothing

DIST_287or387 endp		; End DIST_287or387 procedure
	 subttl  Calculate Maximum Offset
	 page
MAX_OFFSET proc  near
	 assume  ds:nothing,es:nothing,ss:nothing
COMMENT|

On entry:

ES:DI	 ==>	 memory offset

On exit:

ES:DI	 ==>	 same location but with DI as large as possible but
		 not so large as to cause ES to go negative as that
		 would cause us to rely upon the dreaded 1MB wrap.

|

MAXOFF	 equ	 0FFF0h

	 REGSAVE <ax,cx>	; Save registers

	 push	 di		; Save for a moment

	 mov	 cl,4		; Shift amount
	 shr	 di,cl		; Isolate para of control word
	 mov	 cx,es		; Get current segment
	 add	 cx,di		; Add in segment of control word
	 mov	 ax,MAXOFF shr 4 ; Get minimum base

	 sub	 cx,ax		; Less a lot of paras
	 jnc	 short @F	; Jump if not too small

	 add	 ax,cx		; Add it back in get # paras in offset
	 xor	 cx,cx		; Use smallest segment
@@:
	 mov	 es,cx		; Save as segment of control word

	 pop	 di		; Restore

	 mov	 cl,4		; Shift amount
	 shl	 ax,cl		; Convert from paras to bytes
	 and	 di,000Fh	; Isolate offset within byte
	 add	 di,ax		; Include bytes subtracted out above

	 REGREST <cx,ax>	; Restore

	 ret			; Return to caller

	 assume  ds:nothing,es:nothing,ss:nothing

MAX_OFFSET endp 		; End MAX_OFFSET procedure
	 subttl  Convert AX to Hex At ES:DI
	 page
DW2HEX	 proc near
	 assume  ds:nothing,es:PGROUP,ss:nothing
COMMENT|

Convert AX to hex at ES:DI

On entry:

AX	 =	 value to convert
ES:DI	 ==>	 output save area

On exit:

ES:DI	 ==>	 byte to right of converted number

|


	 push	 cx		; Save for a moment
	 mov	 cx,4		; # hex digits
	 call	 BIN2HEX_SUB	; Handle by subroutine
	 pop	 cx		; Restore

	 ret			; Return to caller

	 assume  ds:nothing,es:nothing,ss:nothing

DW2HEX	 endp			; End DW2HEX procedure
	 subttl  Convert EAX to Hex At ES:DI
	 page
DD2HEX	 proc near
	 assume  ds:nothing,es:PGROUP,ss:nothing
COMMENT|

Convert EAX to hex at ES:DI

On entry:

EAX	 =	 value to convert
ES:DI	 ==>	 output save area

On exit:

ES:DI	 ==>	 byte to right of converted number

|


	 push	 cx		; Save for a moment
	 mov	 cx,8		; # hex digits
	 call	 BIN2HEX_SUB	; Handle by subroutine
	 pop	 cx		; Restore

	 ret			; Return to caller

	 assume  ds:nothing,es:nothing,ss:nothing

DD2HEX	 endp			; End DD2HEX procedure
	 subttl  Binary To hex Subroutine
	 page
BIN2HEX_SUB proc near
	 assume  ds:nothing,es:PGROUP,ss:nothing
COMMENT|

Convert AL/AX/EAX to hex at ES:DI

On entry:

AL/AX/EAX=	 value to convert
CX	 =	 # hex digits to convert
ES:DI	 ==>	 output save area

On exit:

ES:DI	 ==>	 byte to right of converted number

|


	 REGSAVE <ax,bx,edx>	; Save registers

	 pushf			; Save flags
	 std			; Store backwards

	 mov	 edx,eax	; Copy to secondary register
	 lea	 bx,HEXTABLE	; XLAT table
	 add	 di,cx		; Skip to the end+1
	 push	 di		; Save to return
	 dec	 di		; Now the last digit
BIN2HEX_MORE:
	 mov	 al,dl		; Copy to XLAT register
	 and	 al,0Fh 	; Isolate low-order digit
	 xlat	 HEXTABLE	; Convert to ASCII hex
	 stosb			; Save in output area

	 shr	 edx,4		; Shift next digit down to low-order

	 loop	 BIN2HEX_MORE	; Jump if more digits to format

	 pop	 di		; Restore

	 popf			; Restore flags

	 REGREST <edx,bx,ax>	; Restore

	 ret			; Return to caller

	 assume  ds:nothing,es:nothing,ss:nothing

BIN2HEX_SUB endp		; End BIN2HEX_SUB procedure
	 subttl  Display CPUID Information
	 page
DISP_CPUID proc  near
	 assume  ds:PGROUP,es:PGROUP,ss:nothing
COMMENT|

Display CPUID information

On entry:

BX	 =	 FLG_REC value

|

	 REGSAVE <eax,dx,di>	; Save registers

	 test	 bx,FLG_CPUID	; Duzit support this instruction?
	 jz	 short DISP_CPUID_EXIT ; Jump if not

; Display manufacturer's ID

	 lea	 dx,MANUID	; DS:DX ==> ASCIIZ string to display
	 mov	 ah,09h 	; Function code to display string at DS:DX
	 int	 21h		; Request DOS service

; Display CPU signature bits

	 mov	 eax,CPUSIG_FLAG ; Get CPU signature bits
	 lea	 di,MSG_CPUSIG1 ; ES:DI ==> output save area
	 call	 DD2HEX 	; Convert EAX to hex at ES:DI
	 lea	 dx,MSG_CPUSIG	; DS:DX ==> ASCIIZ string to display
	 mov	 ah,09h 	; Function code to display string at DS:DX
	 int	 21h		; Request DOS service

; Display CPU feature bits

	 mov	 eax,CPUFET_FLAG ; Get CPU feature bits
	 lea	 di,MSG_CPUFET1 ; ES:DI ==> output save area
	 call	 DD2HEX 	; Convert EAX to hex at ES:DI
	 lea	 dx,MSG_CPUFET	; DS:DX ==> ASCIIZ string to display
	 mov	 ah,09h 	; Function code to display string at DS:DX
	 int	 21h		; Request DOS service

; Display highest CPUID function supported

	 mov	 eax,CPUID_LAST ; Get highest CPUID function supported
	 lea	 di,MSG_CPULST1 ; ES:DI ==> output save area
	 call	 DD2HEX 	; Convert EAX to hex at ES:DI
	 lea	 dx,MSG_CPULST	; DS:DX ==> ASCIIZ string to display
	 mov	 ah,09h 	; Function code to display string at DS:DX
	 int	 21h		; Request DOS service
DISP_CPUID_EXIT:
	 REGREST <di,dx,eax>	; Restore

	 ret			; Return to caller

	 assume  ds:nothing,es:nothing,ss:nothing

DISP_CPUID endp 		; End DISP_CPUID procedure
	 subttl  Display Pentium CPU Information
	 page
DISP_P5  proc	 near
	 assume  ds:PGROUP,es:PGROUP,ss:nothing
COMMENT|

Display Pentium CPU information

On entry:

BX	 =	 FLG_REC value

|

	 REGSAVE <eax,bx,dx,di> ; Save registers

	 and	 bx,mask $FLG_CPU ; Isolate flags

	 cmp	 bx,FLG_P5	; Izit a Pentium CPU?
	 jne	 short DISP_P5_EXIT ; Jump if not

; Display CR4 value

	 mov	 eax,CR4_FLAG	; Get CR4 value
	 lea	 di,MSG_CR4FLG1 ; ES:DI ==> output save area
	 call	 DD2HEX 	; Convert EAX to hex at ES:DI
	 lea	 dx,MSG_CR4FLG	; DS:DX ==> ASCIIZ string to display
	 mov	 ah,09h 	; Function code to display string at DS:DX
	 int	 21h		; Request DOS service
DISP_P5_EXIT:
	 REGREST <di,dx,bx,eax> ; Restore

	 ret			; Return to caller

	 assume  ds:nothing,es:nothing,ss:nothing

DISP_P5  endp			; End DISP_P5 procedure

CODE	 ends			; End CODE segment

	 end	 INITIAL	; End CPUID module
